class << ARGV
  # canonical [= default] [needarg] [short] [:depends ....] [:conflicts ....]
  SpecRegExp = /\s*
		([-\w]+)(?:\s*=\s*(\w+))?(?:\s+:(require|optional|none))?
		(?:\s+(\w+))?
		(?:\s*:depends\s+([^:]+))?
		(?:\s*:conflicts\s+([^:]+))?
		(?:\s*:comment\s+(.+))?/x

  def argspec(spec)
    @options = {}
    @needarg = {}
    @short = {}
    @depends = {}
    @conflicts = {}
    @comment = {}
    spec.each do
      |ln|
      SpecRegExp =~ ln
      @options[$1] = $2 || false
      @needarg[$1] = $3 ? $3.intern : :optional
      if $4
	$4.each_byte do
	  |c|
	  raise "#{c.chr} is used for multiple arguments." if @short[c.chr]
	  @short[c.chr] = $1
	end
      end
      @depends[$1] = $5.split if $5
      @conflicts[$1] = $6.split if $6
      @comment[$1] = $7
    end
  end

  def dumpspec
    @options.keys.sort.each do
      |name|
      puts <<-EOS.gsub(/^\t/, "")
	#{name}[#{@short.select {|k, v| v == name}.collect! {|k, v| k}.sort}] = #{@options[name]} (#{@needarg[name].id2name})
	    depends:\t#{@depends[name] ? @depends[name].join(" ") : "none"}
	    conflicts:\t#{@conflicts[name] ? @conflicts[name].join(" ") : "none"}
	    "#{@comment[name]}"
      EOS
    end
  end

  def parse
    legalopts = @options.keys
    args = []
    while not ARGV.empty?
      arg = ARGV.shift
      case arg
      when /^--([-\w]+)(?:=(.*))?$/
	opt = $1
	val = $2 || true
	if not legalopts.include?(opt)
	  $stderr.puts "unknown option #{opt} is ignored."
	  next
	end
	case @needarg[opt]
	when :none
	  $stderr.puts "#{opt} does not accept argument. ignore #{$2}." if $2
	  @options[opt] = true
	when :require
	  raise ArgumentError, "#{opt} requires argument." unless $2
	  @options[opt] = val
	when :optional
	  @options[opt] = val
	end
      when /^-([\w]+)$/
	opts = $1
	opts.each_byte do
	  |o|
	  opt = @short[o.chr]
	  if not legalopts.include?(opt)
	    $stderr.puts "unknown option #{o.chr} is ignored."
	    next
	  end
	  case @needarg[opt]
	  when :none
	    @options[opt] = true
	  when :require
	    val = ARGV.head
	    if val.nil? or val =~ /^-/
	      raise ArgumentError, "#{opt} requires argument."
	    end
	    ARGV.shift
	    @options[opt] = val
	  when :optional
	    val = ARGV.head
	    if val.nil? or val =~ /^-/
	      @options[opt] = true
	    else
	      ARGV.shift
	      @options[opt] = val
	    end
	  end
	end
      when /^--$/
	break
      else
	args << arg
      end
    end
    while arg = ARGV.shift
      args << arg
    end
    ARGV.replace args

    @options.each do
      |opt, arg|
      next unless arg
      if @depends[opt]
	flag = true
	@depends[opt].each {|dep| flag &&= @options[dep]}
	raise ArgumentError, "#{opt} depends on #{@depends[opt].join(' ')}." unless flag
      end
      if @conflicts[opt]
	flag = false
	@conflicts[opt].each{|conf| flag ||= @options[conf]}
	raise ArgumentError, "#{opt} conflicts with #{@conflicts[opt].join(' ')}." if flag
      end
    end
    @options
  end

  def options
    @options
  end

  def head
    ARGV[0]
  end
end

if $0 == __FILE__

ARGV.argspec <<EOS
  foo :require
  bar b :depends baz :comment this is a comment.
  baz :depends bar
  goo = goo :optional
  car :none :conflicts foo
  caz caz :depends foo :conflicts bar
EOS

ARGV.dumpspec

ARGV.parse.to_a.sort.each {|opt, val| puts "#{opt} = #{val}"}
p ARGV

end
